-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "main"
M.relevantDevice = nil
M.engineInfo = {}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0

-----

local stall_warning_threshold = 10 --AoA at which the stall warning starts to sound

local min = math.min
local max = math.max
local asin = math.asin
local sqrt = math.sqrt
local pi = math.pi

local avToRPM = 9.549296596425384

local engine = nil
local propShaft = nil

local energyStorageData = {
  tanks = {},
  ratio = 0,
  volume = 0,
  volumeAcro = 0, --in the acrobatic (central) tank
  capacity = 0,
  invEnergyStorageCount = 0,
}

local smokeNodes = {}
local stall_warning_e = nil
local hasStallWarning = false

local function updateGFX(dt)
  electrics.values.throttle = min(max(electrics.values.throttle + (electrics.values.throttle_change * dt * 0.5), 0), 1)
  local rpm = engine and engine.outputAV1 * avToRPM or 0

  --Smoke System
  if electrics.values.smokesystem >= 1 and not engine.isDisabled and propShaft and (propShaft.outputAV1 * avToRPM > 100) then
    obj:addParticleByNodesRelative(smokeNodes[2], smokeNodes[1], 25, 2, 0.1, 3)
  end

  local bulletTimeCoef = bullettime.get()
  electrics.values.propellerblur = propShaft and ((propShaft.outputAV1 * avToRPM * bulletTimeCoef) > 600 and 1 or 0) or 0

  energyStorageData.ratio = 0
  energyStorageData.volume = 0
  energyStorageData.volumeAcro = 0

  for _,s in pairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.ratio = energyStorageData.ratio + energyStorage.remainingRatio
      energyStorageData.volume = energyStorageData.volume + energyStorage.remainingVolume
      if s == "acroTank" then energyStorageData.volumeAcro = energyStorage.remainingVolume end
    end
  end
  energyStorageData.ratio = energyStorageData.ratio * energyStorageData.invEnergyStorageCount

  electrics.values.fuel = energyStorageData.ratio or 0
  electrics.values.lowfuel = electrics.values.fuel < 0.1
  electrics.values.fuelCapacity = energyStorageData.invEnergyStorageCount > 0 and (energyStorageData.capacity or 0) or 1
  electrics.values.fuelVolume = energyStorageData.volume
  electrics.values.fuelVolumeAcro = energyStorageData.volumeAcro
  electrics.values.fuelVolumeWings = energyStorageData.volume - energyStorageData.volumeAcro

  electrics.values.brake = min(max(input.parkingbrake, 0), 1)
  electrics.values.clutch = 0
  electrics.values.clutchRatio = 1
  electrics.values.gear = string.format("%d%%", electrics.values.throttle * 100)
  electrics.values.gearIndex = 0
  electrics.values.rpm = rpm
  electrics.values.wheelspeed = electrics.values.airspeed

  if engine then
    M.engineInfo = {
      rpm
      , engine.maxAV * avToRPM
      , 0
      , 0
      , electrics.values.rpm
      , electrics.values.gear
      , 0
      , 0
      , engine.outputTorque1
      , 0
      , obj:getVelocity():length()  -- airspeed
      , electrics.values.fuelVolume
      , electrics.values.fuelCapacity
      , "manual"
      , obj:getID()
      , 0
      , 0
      , 0
      , 0
    }
  end

  --instruments
  electrics.values.heading = -1 * obj:getDirection() * 180/3.14
  electrics.values.accz = sensors.gz2 / obj:getGravity() - 1
  electrics.values.accz_x = max(electrics.values.accz_x, electrics.values.accz)
  electrics.values.accz_n = min(electrics.values.accz_n, electrics.values.accz)

  --Airspeed needele
  --ias = electrics.values.airspeed * 1.94384449
  local ias = (obj:getDirectionVector():dot(obj:getVelocity() - obj:getFlow())) * sqrt(obj:getDensityAtNode(0) / 1.225) * 1.94384449

  --quadratic in the interval 0-56.725 kts, then linear, with a slope change at 180kts
  if ias < 56.725 then
    electrics.values.cust_airspeed = (0.015) * (ias*ias)
  elseif ias <= 180 then
    electrics.values.cust_airspeed = ias * 1.71 - 48.5  --((ias - 40kts) * 240deg / 140kts) + 20deg
  else
    electrics.values.cust_airspeed = ias + 80          --((ias - 180kts) * 20deg / 20kts) + 260deg
  end

  --Joystick Hydros
  electrics.values.jy_fl = - electrics.values.elevator * 0.05 + electrics.values.aileron * 0.1
  electrics.values.jy_fr = - electrics.values.elevator * 0.05 - electrics.values.aileron * 0.1
  electrics.values.jy_bl =   electrics.values.elevator * 0.05 + electrics.values.aileron * 0.1
  electrics.values.jy_br =   electrics.values.elevator * 0.05 - electrics.values.aileron * 0.1

  --Stall Warning (Based on AoA)
  if obj:getAirSpeed() > 5 and hasStallWarning then
    local vv = obj:getVelocity() - obj:getFlow()
    vv:normalize()
    local a = asin(min(1, max(0, (obj:getDirectionVector() - vv).z))) / pi * 180 --angle of attack
    --if a>stall_warning_threshold then
    --electrics.values['stall_warning']=1
    --else
    --stall_warning_e = newExponentialSmoothing(5)
    --electrics.values['stall_warning']=0
    --end
    electrics.values.stall_warning = min(1, max(0, stall_warning_e:get((a - stall_warning_threshold)*0.6)))
    if electrics.values.stall_warning < 0 then
      stall_warning_e:reset()
    end
  else
    electrics.values.stall_warning = 0
  end

end

local function init(jbeamData)
  engine = powertrain.getDevice("engine")
  propShaft = powertrain.getDevice("shaft")

  smokeNodes = {}
  for k, node in pairs (v.data.nodes) do
    if node.name == "enc1" then smokeNodes[1] = k  end
    if node.name == "enc2" then smokeNodes[2] = k   end
  end

  electrics.values.throttle = 0
  electrics.values.throttle_change = 0
  electrics.values.smokesystem = 0
  electrics.values.propellerblur = 0

  electrics.values.jy_fl = 0
  electrics.values.jy_fr = 0
  electrics.values.jy_bl = 0
  electrics.values.jy_br = 0

  electrics.values.stall_warning = 0
  stall_warning_e = newExponentialSmoothing(10)
  hasStallWarning = jbeamData.hasStallWarning or false

  electrics.values.heading = 0
  electrics.values.accz = 0
  electrics.values.accz_x = 0
  electrics.values.accz_n = 0

  energyStorageData = {
    tanks = {},
    ratio = 0,
    volume = 0,
    volumeAcro = 0, --in the acrobatic (central) tank
    capacity = 0,
    invEnergyStorageCount = 0,
  }

  if engine then
    for _,s in pairs(engine.energyStorage or {}) do
      table.insert(energyStorageData.tanks, s)
    end
  end

  local energyStorageCount = 0
  for _,s in ipairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.capacity = energyStorageData.capacity + energyStorage.capacity
      energyStorageCount = energyStorageCount + 1
    end
  end

  energyStorageData.invEnergyStorageCount = energyStorageCount > 0 and 1 / energyStorageCount or 0
end

local function sendTorqueData()
  if not playerInfo.firstPlayerSeated then return end
  if engine then
    engine:sendTorqueData()
  end
end

M.init = init
M.updateGFX = updateGFX

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setGearboxMode = nop
M.setStarter = nop
M.freeze = nop
M.sendTorqueData = sendTorqueData
M.setEngineIgnition = nop

-------------------------------


return M